#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>

const double PI = 3.14159265359;

double f(double, double);

int main(int argc, char ** argv)
{
  if(argc != 5)
  {
    printf("Invalid number of arguments.\n");
    exit(0);
  }

  double h = 0.0001; //The rectangle width
  double startX = atof(argv[1]), endX = atof(argv[2]);  //The starting and ending points
  double startY = atof(argv[3]), endY = atof(argv[4]);  //The starting and ending points

  if(startX > endX)
  {
    printf("Invalid value for starting X.\n");
    exit(0);
  }

  if(startY > endY)
  {
    printf("Invalid value for starting Y.\n");
    exit(0);
  }

  double x = startX; //The current value of x.
  double y = startY; //The current value of y.
  double volume = 0.0;

  int rank, size;
  long i, j,
    endXdir = (long) ( (endX - startX) / h ),
    endYdir = (long) ( (endY - startY) / h );


  MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    long workToDo = endXdir / size;
    long loopStart = rank*workToDo;
    long workToDoY = endYdir;

    if(rank == size-1 && endXdir % size != 0)
      workToDo += endXdir%size;

    #pragma omp parallel for private(i,j) shared(endYdir, loopStart, workToDo, x, y, h) reduction(+:volume)
    for(i = loopStart; i < loopStart+workToDo ; i++)
      for(j = 0; j < endYdir; j++)
        volume += h*h*f(x + i*h, y + j*h);

    if(rank == 0)
    {
      double additionalVolume;
      for(i = 1; i < size; i++)
      {
        MPI_Recv(&additionalVolume, 1, MPI_DOUBLE, i,0,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
        volume += additionalVolume;
      }
      printf("The volume is %lf\n", volume);
    } else {
      MPI_Send(&volume, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
    }

  MPI_Finalize();
  return 0;
}

double f(double x, double y)
{
  return -cos(x)*sin(y)*exp( -( (x - PI)*(x - PI) + (y - PI)*(y - PI) ) );
}
