#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>

#define X 100

int main(int argc, char ** argv)
{
  int rank, npes, i;
  int * randomInts;
  long totalSum;
  int workToDo;

  int N;

  if(argc != 2)
  {
    printf("Incorrect number of arguments.\n");
    exit(0);
  }

  //Read the value
  N = atof(argv[1]);

  if(N <= 0)
  {
    printf("Invalid value for array size.\n");
    exit(0);
  }

  MPI_Init(&argc, &argv);

  //Get rank and number of processes
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &npes);

  //Generate N random integers valued between 0 and X-1
  if(rank == 0)
  {
    randomInts = (int *) malloc(N*sizeof(int));
    for(i = 0; i < N; i++)
      randomInts[i] = rand()%X;

    workToDo = N/npes;
  }

  //Technically not necessary, but used so all concepts are covered.
  MPI_Bcast(&workToDo, 1, MPI_INT, 0, MPI_COMM_WORLD);

  //Scatter the integers to each of npes processes
  int * workArray = (int *) malloc(workToDo*sizeof(int));
  MPI_Scatter(randomInts, workToDo, MPI_INT, workArray, workToDo, MPI_INT, 0, MPI_COMM_WORLD);

  //Compute the sum locally.
  long sum = 0L;
  for(i = 0; i < workToDo; i++)
    sum += workArray[i];

  //Take care of extra integers
  if(rank == 0)
    if(workToDo * npes < N)
      for(i = workToDo*npes; i < N; i++)
        sum += randomInts[i];

  MPI_Reduce(&sum, &totalSum, 1, MPI_LONG, MPI_SUM, 0, MPI_COMM_WORLD);

  free((void *) workArray);

  if(rank == 0)
  {
    printf("The total sum is %ld\n", totalSum);

    //Error check
    int j;
    sum = 0;
    for(i = 0; i < N; i++)
        sum += randomInts[i];
    
    printf("The actual sum in process %d is %ld\n", 0, sum);
    free((void *)randomInts);
  }

  MPI_Finalize();
  return 0;
}
