#include <stdio.h>
#include <mpi.h>
// #include <omp.h>
#include <math.h>
#include <stdlib.h>

#define NUM_ITER 200000000

double nextDouble(double, double);
double mineshaft(double);

int main(int argc, char ** argv)
{
  double startX, endX;
  double globalMin;
  double locations[2];
  int rank, npes;
  
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &npes);


  if(rank == 0)
  {
    if(argc != 3)
    {
      printf("Invalid number of arguments, provide starting and ending x values.\n");
      exit(0);
    }

    locations[0] = atof(argv[1]);
    locations[1] = atof(argv[2]);

    if(locations[0] > locations[1])
    {
      printf("Invalid starting x value, provide valid starting and ending x values.\n");
      exit(0);
    }
  }


  MPI_Bcast(locations, 2, MPI_DOUBLE, 0, MPI_COMM_WORLD);

  startX = locations[0];
  endX = locations[1];


  long workToDo = NUM_ITER/npes;
  long i;
  
  srand(1000000*rank);

  double startTime, endTime;
  if(rank == 0)
    startTime = MPI_Wtime();

  double minLoc = nextDouble(startX, endX);
  double min = mineshaft(minLoc); 
  double newMinLoc, newMin;

  #pragma omp parallel for private(newMin, i) shared(startX, endX, workToDo) reduction(min:min)
  for(i = 0; i < workToDo; i++)
  {
    newMinLoc = nextDouble(startX, endX);
    newMin = mineshaft(newMinLoc);
    if(newMin < min)
    {
      min = newMin;
      minLoc = newMinLoc;
    }
  }
//  printf("min = %lf location = %lf in process %d\n", min, minLoc, rank);
  double result[2];
  double * finalResults;

  result[0] = min;
  result[1] = minLoc;

  if(rank == 0)
  {
    finalResults = (double *) malloc(sizeof(double)*npes*2);
  }
  
  MPI_Gather(result, 2, MPI_DOUBLE, finalResults, 2, MPI_DOUBLE, 0, MPI_COMM_WORLD);

  if(rank == 0)
  {
    endTime = MPI_Wtime();
    int minLoc = 0;
    for(i = 2; i < npes*2; i+=2)
    {
      if(finalResults[i] < finalResults[minLoc])
        minLoc = i;
    }
    printf("The global minimum is %lf\n", finalResults[minLoc]);
    printf("The location of the global minimum is %lf\n", finalResults[minLoc+1]);
    free((void *) finalResults);
    printf("The true minimum is %lf at %lf\n", mineshaft(5.0), 5.0);
    printf("The time to complete the computations was %lf s\n", endTime - startTime);
  }
    
  MPI_Finalize();

  return 0;
}

double nextDouble(double start, double end)
{
  return (rand()/(double)RAND_MAX) * (end - start) + start;
}

double mineshaft(double x)
{
  double result = cos(x)+pow(pow(7.0-x, 2.0),1.0/15.0)+2.0*pow(pow(5.0-x,4.0),1.0/35.0);
  return result;
}
